
/* S2 cpu micro architecture simulator

   5 December 2001
   s21  18 Jan 2007

   modify from s1 sim

   Computer Architecture 1997, 1998

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University

   S2 calculator				9 Feb 2007
*/

#include "s21.h"

#define MXBUF	500
#define sep 	" \n"		// space is a separator
#define jtab    132         // first entry of jump table

int R[32], Ir, Mdr, Mar, Pc, T, Z, S, M[20000];
int runflag, savePc;
char inbuf[MXBUF], *cp;
int type, value;
FILE *fi;

extern int ip;

void ifetch(void){      // instruction fetch
	Mar = Pc;
	Mdr = M[Mar]; Pc = Pc + 1;
}

void decode(void){
	Ir = Mdr;
}

/* alu operation input : a, b
 affect condition code Z,S
 carry and overflow are not implemented */

int alu(char op, int a, int b){
  int t;
  switch(op) {
  case '+': t = a + b; break;
  case '-': t = a - b; break;
  case '*': t = a * b; break;
  case '/': t = a / b; break;
  case '&': t = a & b; break;
  case '|': t = a | b; break;
  case '^': t = a ^ b; break;
  case '<': t = a << b; break;
  case '>': t = a >> b; break;
  case 'E': t = a == b; break;
  case 'e': t = a != b; break;
  case 'L': t = a < b; break;
  case 'l': t = a <= b; break;
  case 'G': t = a > b; break;
  case 'g': t = a >= b; break;
  }
  if( t == 0 ) Z = 1; else Z = 0;
  if( t < 0 )  S = 1; else S = 0;
  return t;
}

int adder(int a, int b){
	return a + b;
}
/*
int testCc(int cond ){	// test condition code
  switch( cond ){
  case ALWAYS: return 1;
  case EQ: return Z;
  case NEQ: return Z == 0;
  case LT: return S && (Z == 0);
  case LE: return S || Z;
  case GE: return (S == 0) || Z;
  case GT: return (S == 0) && (Z == 0);
  default: error("undefine conditional code");
  }
  return 0;
}
*/
// write register, check for R[0]
void wR(int rn, int value){
	if(rn != 0) R[rn] = value;
}

void opRR(char op){
	T = alu(op,R[IRr2()],R[IRr3()]);
	wR(IRr1(),T);
}

void opRI(char op){
	T = alu(op,R[IRr2()],IRdisp());
	wR(IRr1(),T);
}

PRIVATE int signx2( int d ){	// sign bit 22 extended
	if( d & 0x0400000 ) return d | 0xFFC00000;
	return d;
}

void run(void){		// execute one instruction
	savePc = Pc;
	ifetch();
	decode();
//	countclock();
//	ninst++;
	switch( IRop() ) {
	case LDA:
		Mar = IRads();
		Mdr = M[Mar];
		wR(IRr1(),Mdr);
		break;
	case LDD:
//		T = alu('+',R[IRr2()],IRdisp());
//		Mar = T;
		Mar = adder(R[IRr2()],IRdisp());  // not affect flag
		Mdr = M[Mar];
		wR(IRr1(),Mdr);
		break;
	case STA:
		Mar = IRads();
		Mdr = R[IRr1()];
		M[Mar] = Mdr;
		break;
	case STD:
//		T = alu('+',R[IRr2()],IRdisp());
//		Mar = T;
		Mar = adder(R[IRr2()],IRdisp());
		Mdr = R[IRr1()];
		M[Mar] = Mdr;
		break;
	case MVI:
		wR(IRr1(),signx2(IRads()));
		break;
	case JMP:
//		if( testCc(IRr1()) )
			Pc = IRads();
		break;
	case JAL:
		wR(IRr1(),Pc);
		Pc = IRads();
		break;
	case JT:
		if( R[IRr1()] != 0 ) Pc = IRads();
		break;
	case JF:
		if( R[IRr1()] == 0 ) Pc = IRads();
		break;
	case ADDI: opRI('+'); break;
	case SUBI: opRI('-'); break;
	case MULI: opRI('*'); break;
	case DIVI: opRI('/'); break;
	case ANDI: opRI('&'); break;
	case ORI:  opRI('|'); break;
	case XORI: opRI('^'); break;
	case EQI: opRI('E'); break;
	case NEI: opRI('e'); break;
	case LTI: opRI('L'); break;
	case LEI: opRI('l'); break;
	case GTI: opRI('G'); break;
	case GEI:  opRI('g'); break;
	case SHLI: opRI('<'); break;
	case SHRI: opRI('>'); break;
	case XOP: xop(); break;   /* extended op */
	default: error("undefine op");
	}
}

void xop(void){
	switch( IRxop() ) {
	case ADD: opRR('+'); break;
	case SUB: opRR('-'); break;
	case MUL: opRR('*'); break;
	case DIV: opRR('/'); break;
	case AND: opRR('&'); break;
	case OR:  opRR('|'); break;
	case XOR: opRR('^'); break;
	case EQ: opRR('E'); break;
	case NE: opRR('e'); break;
	case LT: opRR('L'); break;
	case LE: opRR('l'); break;
	case GT: opRR('G'); break;
	case GE:  opRR('g'); break;
	case SHL: opRR('<'); break;
	case SHR: opRR('>'); break;
	case MV: wR(IRr1(),R[IRr2()]); break;
	case LDX:
//		T = alu('+',R[IRr2()],R[IRr3()]);
//		Mar = T;
		Mar = adder(R[IRr2()],R[IRr3()]); // not affect flag
		Mdr = M[Mar];
		wR(IRr1(),Mdr);
		break;
	case STX:
//		T = alu('+',R[IRr2()],R[IRr3()]);
//		Mar = T;
		Mar = adder(R[IRr2()],R[IRr3()]);
		Mdr = R[IRr1()];
		M[Mar] = Mdr;
		break;
	case RET: Pc = R[IRr1()]; break;
	case TRAP: trap(IRr1()); break;
	case PUSH:
		wR(IRr1(),R[IRr1()]+1);
		Mar = R[IRr1()];
		Mdr = R[IRr2()];
		M[Mar] = Mdr;
		break;
	case POP:
		Mar = R[IRr1()];
		Mdr = M[Mar];
		wR(IRr2(),Mdr);
		wR(IRr1(),R[IRr1()]-1);
		break;
	case NOT:
		if(R[IRr2()] != 0) wR(R[IRr1()],0);
		else wR(R[IRr1()],1);
		break;
	default: error("undefine op");
	}
}

// get one token from input
char *tok(void){
	if( cp != NULL ) cp = strtok(NULL,sep);
	while ( cp == NULL ){
		printf(">");
		cp = fgets(inbuf,MXBUF,fi);
		if( cp != NULL ) cp = strtok(inbuf, sep);
	}
	return cp;
}

// token of calculator (keys)
char token[][4] = {
	"+","-","*","/","f0","f1","f2","f3","f4","f5",
	"f6","f7","f8","f9","=","ld"
};

void loadfun(char *fn){
	int i, n, start, k;
	char nm[80];

	n = strcspn(fn,".");
	strncpy(nm,fn,n);
	nm[n] = 0;
	for(i=4; i<=13; i++)	// check f0..f9
		if(eqs(nm,token[i])) break;
	if(i > 13) error("not valid function object");
	i -= 4;			// now i = 0..9
	n = jtab + i;	// corresponding entry of jump table
	start = ip;		// current CS
	loadprogram(fn);
	for(i=1000, k=start; i<ip; i++, k++)
		M[k] = M[i];
	ip = k;
	M[n] = (M[n] & 0xF8000000) | start;  // patch entry
}

void trap( int num ){
	int i, n;

	switch( num ) {
	case 0 :  // stop
		runflag = 0;
		break;
	case 1 : // print integer
		printf("%d\n", R[30]);
		break;
	case 2 : // printc
		printf("%c", R[30]);
		break;
	case 3 : // read a token from inbuf[];
		cp = tok();
		if(cp == NULL){				// end of line
			type = 18;
		}else if( strcspn(cp,"0123456789") == 0){  // number
			value = atoi(cp);
			M[901] = value;
			type = 1;
		}else{
			for(i=0; i<=15; i++)
				if(eqs(cp,token[i])) break;
			type = i + 2;
			if(type > 17) error("unknown token");
		}
		R[28] = type;
		break;
	case 4 : // ld file.obj
		cp = tok();
		loadfun(cp);
		break;
	}
}

void mainCal(void){
	fi = stdin;
	while(1){
		runflag = 1;
		Pc = 0;
		while(runflag){
			run(); // show();
		}
	}
}

void main(int argc, char *argv[]){
/*
	if( argc < 2 ) {
		printf("usage : s2 objfile\n");
		exit(0);
	}
	loadprogram(argv[1]);
*/

	loadprogram("s2cal.obj");
//	clock = 0;
//	ninst = 0;
	Pc = 0;
	Z = 0;
	S = 0;
	R[0] = 0;
	runflag = 1;
	R[27] = 3000;   // eval stack
	R[31] = 2000;   // system stack
	mainCal();
}

